/* Just a simple example program that creates a terminal in a frame
 * and lets the user interact with it.
 *
 * To compile:
 *    gcc -o boxshell boxshell.c $(pkg-config madtty --cflags --libs)
 */

#include <ncurses.h>
#include <locale.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include <madtty/madtty.h>

static unsigned char getout = 0;
static int screen_w, screen_h;
static WINDOW *term_win;

void sigchld(int signo __attribute__((unused)))
{
    getout = 1;
}

int main(int argc, char *argv[])
{
    struct timeval next = { 0, 0 };
    RoteTerm *rt;
    int i, j, ch, w, h, pos;
    char buf[BUFSIZ];

    signal(SIGCHLD, sigchld);

    w = 80;
    h = 40;
    if (argc > 1) {
        char *p = argv[1];
        w = strtol(p, &p, 10);
        if (*p++ == 'x')
            h = strtol(p, &p, 10);
    }

    setlocale(LC_ALL, "");
    initscr();
    noecho();
    start_color();
    raw();
    nodelay(stdscr, TRUE);       /* prevents getch() from blocking; rather
                                  * it will return ERR when there is no
                                  * keypress available */

    keypad(stdscr, TRUE);        /* necessary to use rote_vt_keypress */
    getmaxyx(stdscr, screen_h, screen_w);

    /* initialize the color pairs the way rote_vt_draw expects it. You might
     * initialize them differently, but in that case you would need
     * to supply a custom conversion function for rote_vt_draw to
     * call when setting attributes. The idea of this "default" mapping
     * is to map (fg, bg) to the color pair bg * 8 + 7 - fg. This way,
     * the pair (white, black) ends up mapped to 0, which means that
     * it does not need a color pair (since it is the default). Since
     * there are only 63 available color pairs (and 64 possible fg/bg
     * combinations), we really have to save 1 pair by assigning no pair
     * to the combination white/black. */
    for (i = 0; i < 8; i++) for (j = 0; j < 8; j++)
        if (i != 7 || j != 0)
            init_pair(j*8+7-i, i, j);

    /* paint the screen blue */
    attrset(COLOR_PAIR(32));
    for (i = 0; i < screen_h; i++)
        for (j = 0; j < screen_w; j++)
            addch(' ');
    refresh();

    /* create a window with a frame */
    term_win = newwin(h + 2, w + 2, 2, 3);
    wborder(term_win, '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0');
    mvwprintw(term_win, 0, 27, " Term In a Box ");
    wrefresh(term_win);

    rt = rote_vt_create(h, w);
    {
        const char *path = getenv("SHELL") ?: "/bin/sh";
        const char *args[] = { path, "--login", NULL};

        rote_vt_forkpty(rt, path, args);
    }

    /* keep reading keypresses from the user and passing them to the terminal;
     * also, redraw the terminal to the window at each iteration */
    ch = '\0';
    pos = 0;
    while (!getout) {
        fd_set rfds;
        struct timeval tv = { 0 , 1000 }, t;

        FD_ZERO(&rfds);
        FD_SET(rt->pty, &rfds);

        if (select(rt->pty + 1, &rfds, NULL, NULL, &tv) > 0) {
            int nb;

            nb = rote_vt_read(rt, buf + pos, sizeof(buf) - pos);
            if (nb <= 0)
                continue;
            pos += nb;

            nb = rote_vt_inject(rt, buf, pos);
            if (nb <= 0)
                continue;
            memmove(buf, buf + nb, pos - nb);
            pos -= nb;
        }

        while ((ch = getch()) != ERR) {
            rote_vt_keypress(rt, ch); /* pass the keypress for handling */
        }

        gettimeofday(&t, NULL);
        if (timercmp(&t, &next, >=)) {
            rote_vt_draw(rt, term_win, 1, 1);
            wrefresh(term_win);
            gettimeofday(&next, NULL);
            next.tv_usec += 1000 * 1000 / 100;
            if (next.tv_usec > 1000 * 1000) {
                next.tv_usec -= 1000 * 1000;
                next.tv_sec++;
            }
        }
    }

    endwin();
    return 0;
}
