/*
    LICENSE INFORMATION:
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License (LGPL) as published by the Free Software Foundation.

    Please refer to the COPYING file for more information.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    Copyright © 2004 Bruno T. C. de Oliveira
    Copyright © 2006 Pierre Habouzit
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <pty.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include "madtty.h"

#define MAX_CSI_ES_PARAMS 32
#define ESEQ_BUF_SIZE 128  /* size of escape sequence buffer */

/* Terminal private data */
struct RoteTermPrivate_ {
    unsigned escaped : 1;      /* whether we are currently reading an
                                * escape sequence */

    unsigned graphmode : 1;    /* whether terminal is in graphical 
                                * character mode or not */

    int scrolltop, scrollbottom;  /* current scrolling region of terminal */
    int saved_x, saved_y;         /* saved cursor position */

    char esbuf[ESEQ_BUF_SIZE]; /* 0-terminated string. Does NOT include
                                * the initial escape (\x1B) character. */
    int esbuf_len;             /* length of buffer. The following property
                                * is always kept: esbuf[esbuf_len] == '\0' */
};

static void clamp_cursor_to_bounds(RoteTerm *rt)
{
    if (rt->crow < 0) {
        rt->curpos_dirty = true;
        rt->crow = 0;
    }
    if (rt->ccol < 0) {
        rt->curpos_dirty = true;
        rt->ccol = 0;
    }

    if (rt->crow >= rt->rows) {
        rt->curpos_dirty = true;
        rt->crow = rt->rows - 1;
    }
    if (rt->ccol >= rt->cols) {
        rt->curpos_dirty = true;
        rt->ccol = rt->cols - 1;
    }
}

static void cursor_line_down(RoteTerm *rt)
{
    int i;

    rt->crow++;
    rt->curpos_dirty = true;
    if (rt->crow <= rt->pd->scrollbottom)
        return;

    /* must scroll the scrolling region up by 1 line, and put cursor on
     * last line of it */
    rt->crow = rt->pd->scrollbottom;

    for (i = rt->pd->scrolltop; i < rt->pd->scrollbottom; i++) {
        rt->line_dirty[i] = true;
        memcpy(rt->cells[i], rt->cells[i+1], sizeof(RoteCell) * rt->cols);
    }

    rt->line_dirty[rt->pd->scrollbottom] = true;

    /* clear last row of the scrolling region */
    for (i = 0; i < rt->cols; i++) {
        rt->cells[rt->pd->scrollbottom][i].s[0] = 0x20;
        rt->cells[rt->pd->scrollbottom][i].len  = 1;
        rt->cells[rt->pd->scrollbottom][i].attr = 0x70;
    }
}

static void cursor_line_up(RoteTerm *rt)
{
    int i;

    rt->crow--;
    rt->curpos_dirty = true;
    if (rt->crow >= rt->pd->scrolltop)
        return;

    /* must scroll the scrolling region up by 1 line, and put cursor on
     * first line of it */
    rt->crow = rt->pd->scrolltop;

    for (i = rt->pd->scrollbottom; i > rt->pd->scrolltop; i--) {
        rt->line_dirty[i] = true;
        memcpy(rt->cells[i], rt->cells[i-1], sizeof(RoteCell) * rt->cols);
    }

    rt->line_dirty[rt->pd->scrolltop] = true;

    /* clear first row of the scrolling region */
    for (i = 0; i < rt->cols; i++) {
        rt->cells[rt->pd->scrolltop][i].s[0] = 0x20;
        rt->cells[rt->pd->scrolltop][i].len  = 1;
        rt->cells[rt->pd->scrolltop][i].attr = 0x70;
    }
}

static void put_normal_char(RoteTerm *rt, const char *s, int len)
{
    if (rt->ccol >= rt->cols) {
        rt->ccol = 0;
        cursor_line_down(rt);
    }

    if (rt->insert) {
        int i;

        for (i = rt->cols - 1; i >= rt->ccol+1; i--) {
            rt->cells[rt->crow][i] = rt->cells[rt->crow][i-1];
        }
    }

    memcpy(rt->cells[rt->crow][rt->ccol].s, s, len);
    rt->cells[rt->crow][rt->ccol].len  = len;
    rt->cells[rt->crow][rt->ccol].attr = rt->curattr;
    rt->ccol++;

    rt->line_dirty[rt->crow] = true;
    rt->curpos_dirty = true;
}

static void put_graphmode_char(RoteTerm *rt, int c)
{
    char nc;
    /* do some very pitiful translation to regular ascii chars */
    switch (c) {
      case 'j': case 'k': case 'l': case 'm': case 'n': case 't':
      case 'u': case 'v': case 'w':
        nc = '+'; break;
      case 'x':
        nc = '|'; break;
      default:
        nc = '%';
    }

    put_normal_char(rt, &nc, 1);
}

static void new_escape_sequence(RoteTerm *rt)
{
    rt->pd->escaped = true;
    rt->pd->esbuf_len = 0;
    rt->pd->esbuf[0] = '\0';
}

static void cancel_escape_sequence(RoteTerm *rt)
{
    rt->pd->escaped = false;
    rt->pd->esbuf_len = 0;
    rt->pd->esbuf[0] = '\0';
}

static void handle_control_char(RoteTerm *rt, int c)
{
    switch (c) {
      case '\r':  /* carriage return */
        rt->ccol = 0;
        break;

      case '\n':  /* line feed */
        rt->ccol = 0;
        cursor_line_down(rt);
        rt->curpos_dirty = true;
        break;

      case '\b': /* backspace */
        if (rt->ccol > 0)
            rt->ccol--;
        rt->curpos_dirty = true;
        break;

      case '\t': /* tab */
        rt->ccol = (rt->ccol + 8) & ~7;
        clamp_cursor_to_bounds(rt);
        break;

      case '\x1b': /* begin escape sequence (aborting previous one if any) */
        new_escape_sequence(rt);
        break;

      case '\x0e': /* enter graphical character mode */
        rt->pd->graphmode = true;
        break;

      case '\x0f': /* exit graphical character mode */
        rt->pd->graphmode = false;
        break;

      case '\x9b': /* CSI character. Equivalent to ESC [ */
        new_escape_sequence(rt);
        rt->pd->esbuf[rt->pd->esbuf_len++] = '[';
        break;

      case '\x18':
      case '\x1a': /* these interrupt escape sequences */
        cancel_escape_sequence(rt);
        break;

      case '\a': /* bell */
        /* do nothing for now... maybe a visual bell would be nice? */
        break;
    }
}

static bool is_valid_csi_ender(int c)
{
    return (c >= 'a' && c <= 'z')
        || (c >= 'A' && c <= 'Z')
        || (c == '@' || c == '`');
}

static void rote_es_interpret_csi(RoteTerm *rt);

static void try_interpret_escape_seq(RoteTerm *rt)
{
    char firstchar = rt->pd->esbuf[0];
    char lastchar  = rt->pd->esbuf[rt->pd->esbuf_len-1];

    if (!firstchar)
        return;  /* too early to do anything */

    /* interpret ESC-M as reverse line-feed */
    if (firstchar == 'M') {
        cursor_line_up(rt);
        cancel_escape_sequence(rt);
        return;
    }

    if (firstchar != '[' && firstchar != ']') {
        /* unrecognized escape sequence. Let's forget about it. */
        cancel_escape_sequence(rt);
        return;
    }

    if (firstchar == '[' && is_valid_csi_ender(lastchar)) {
        /* we have a csi escape sequence: interpret it */
        rote_es_interpret_csi(rt);
        cancel_escape_sequence(rt);
    } else if (firstchar == ']' && lastchar == '\a') {
        /* we have an xterm escape sequence: interpret it */

        /* rote_es_interpret_xterm_es(rt);     -- TODO!*/
        cancel_escape_sequence(rt);
    }

    /* if the escape sequence took up all available space and could
     * not yet be parsed, abort it */
    if (rt->pd->esbuf_len + 1 >= ESEQ_BUF_SIZE)
        cancel_escape_sequence(rt);
}

int rote_vt_inject(RoteTerm *rt, const char *data, int len)
{
    int pos;

    for (pos = 0; pos < len; pos++) {
        if ((unsigned char)data[pos] <= 31) {
            handle_control_char(rt, data[pos]);
            continue;
        }

        if (rt->pd->escaped && rt->pd->esbuf_len < ESEQ_BUF_SIZE) {
            /* append character to ongoing escape sequence */
            rt->pd->esbuf[rt->pd->esbuf_len] = data[pos];
            rt->pd->esbuf[++rt->pd->esbuf_len] = 0;

            try_interpret_escape_seq(rt);
        } else
        if (rt->pd->graphmode) {
            put_graphmode_char(rt, data[pos]);
        } else {
            static int8_t const lens[5] = { 1, -1, 2, 3, 4 };
            int bsf = __builtin_clz(~((unsigned char)data[pos] << 24));

            if (pos + lens[bsf] > len)
                return pos;

            put_normal_char(rt, data + pos, lens[bsf]);
            pos += lens[bsf] - 1;
        }
    }

    return len;
}

/****************************************************************************/
/* CSI things                                                               */
/****************************************************************************/

/* interprets a 'set attribute' (SGR) CSI escape sequence */
static void interpret_csi_SGR(RoteTerm *rt, int param[], int pcount)
{
    int i;

    if (pcount == 0) {
        /* special case: reset attributes */
        rt->curattr = 0x70;
        return;
    }

    for (i = 0; i < pcount; i++) {

        // From http://vt100.net/docs/vt510-rm/SGR table 5-16
        // 0 	All attributes off
        // 1 	Bold
        // 4 	Underline
        // 5 	Blinking
        // 7 	Negative image
        // 8 	Invisible image
        // 10 	The ASCII character set is the current 7-bit
        //	display character set (default) - SCO Console only.
        // 11 	Map Hex 00-7F of the PC character set codes
        //	to the current 7-bit display character set
        //	- SCO Console only.
        // 12 	Map Hex 80-FF of the current character set to
        //	the current 7-bit display character set - SCO
        //	Console only.
        // 22 	Bold off
        // 24 	Underline off
        // 25 	Blinking off
        // 27 	Negative image off
        // 28 	Invisible image off

        if (param[i] == 0) rt->curattr = 0x70;
        else if (param[i] == 1 || param[i] == 2 || param[i] == 4)  /* set bold */
            ROTE_ATTR_MOD_BOLD(rt->curattr,1);
        else if (param[i] == 5)  /* set blink */
            ROTE_ATTR_MOD_BLINK(rt->curattr,1);
        else if (param[i] == 7 || param[i] == 27) { /* reverse video */
            int fg = ROTE_ATTR_FG(rt->curattr);
            int bg = ROTE_ATTR_BG(rt->curattr);
            ROTE_ATTR_MOD_FG(rt->curattr, bg);
            ROTE_ATTR_MOD_BG(rt->curattr, fg);
        }
        else if (param[i] == 8) rt->curattr = 0x0;    /* invisible */
        else if (param[i] == 22 || param[i] == 24) /* bold off */
            ROTE_ATTR_MOD_BOLD(rt->curattr,0);
        else if (param[i] == 25) /* blink off */
            ROTE_ATTR_MOD_BLINK(rt->curattr,0);
        else if (param[i] == 28) /* invisible off */
            rt->curattr = 0x70;
        else if (param[i] >= 30 && param[i] <= 37)    /* set fg */
            ROTE_ATTR_MOD_FG(rt->curattr, param[i] - 30);
        else if (param[i] >= 40 && param[i] <= 47)    /* set bg */
            ROTE_ATTR_MOD_BG(rt->curattr, param[i] - 40);
        else if (param[i] == 39)  /* reset foreground to default */
            ROTE_ATTR_MOD_FG(rt->curattr, 7);
        else if (param[i] == 49)  /* reset background to default */
            ROTE_ATTR_MOD_BG(rt->curattr, 0);
    }
}

/* interprets an 'erase display' (ED) escape sequence */
static void interpret_csi_ED(RoteTerm *rt, int param[], int pcount)
{
    int r, c;
    int start_row, start_col, end_row, end_col;

    /* decide range */
    if (pcount && param[0] == 2) {
        start_row = 0;
        start_col = 0;
        end_row = rt->rows - 1;
        end_col = rt->cols - 1;
    } else
    if (pcount && param[0] == 1) {
        start_row = 0;
        start_col = 0;
        end_row = rt->crow;
        end_col = rt->ccol;
    } else {
        start_row = rt->crow;
        start_col = rt->ccol;
        end_row = rt->rows - 1;
        end_col = rt->cols - 1;
    }

    /* clean range */
    for (r = start_row; r <= end_row; r++) {
        rt->line_dirty[r] = true;

        for (c = (r == start_row ? start_col : 0);
             c <= (r == end_row ? end_col : rt->cols - 1);
             c++)
        {
            rt->cells[r][c].s[0] = 0x20;
            rt->cells[r][c].len  = 1;
            rt->cells[r][c].attr = rt->curattr;
        }
    }
}

/* interprets a 'move cursor' (CUP) escape sequence */
static void interpret_csi_CUP(RoteTerm *rt, int param[], int pcount)
{
    if (pcount == 0) {
        /* special case */
        rt->crow = rt->ccol = 0;
        return;
    } else
    if (pcount < 2) {
        return;  /* malformed */
    }

    rt->crow = param[0] - 1;  /* convert from 1-based to 0-based */
    rt->ccol = param[1] - 1;  /* convert from 1-based to 0-based */

    rt->curpos_dirty = true;

    clamp_cursor_to_bounds(rt);
}

/* Interpret the 'relative mode' sequences: CUU, CUD, CUF, CUB, CNL,
 * CPL, CHA, HPR, VPA, VPR, HPA */
static void interpret_csi_C(RoteTerm *rt, char verb, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;

    switch (verb) {
      case 'A':           rt->crow -= n; break;
      case 'B': case 'e': rt->crow += n; break;
      case 'C': case 'a': rt->ccol += n; break;
      case 'D':           rt->ccol -= n; break;
      case 'E':           rt->crow += n; rt->ccol = 0; break;
      case 'F':           rt->crow -= n; rt->ccol = 0; break;
      case 'G': case '`': rt->ccol  = param[0] - 1; break;
      case 'd':           rt->crow  = param[0] - 1; break;
    }

    rt->curpos_dirty = true;
    clamp_cursor_to_bounds(rt);
}

/* Interpret the 'erase line' escape sequence */
static void interpret_csi_EL(RoteTerm *rt, int param[], int pcount)
{
    int erase_start, erase_end, i;
    int cmd = pcount ? param[0] : 0;

    switch (cmd) {
      case 1:  erase_start = 0;        erase_end = rt->ccol;     break;
      case 2:  erase_start = 0;        erase_end = rt->cols - 1; break;
      default: erase_start = rt->ccol; erase_end = rt->cols - 1; break;
    }

    for (i = erase_start; i <= erase_end; i++) {
        rt->cells[rt->crow][i].s[0] = 0x20;
        rt->cells[rt->crow][i].len  = 1;
        rt->cells[rt->crow][i].attr = rt->curattr;
    }

    rt->line_dirty[rt->crow] = true;
}

/* Interpret the 'insert blanks' sequence (ICH) */
static void interpret_csi_ICH(RoteTerm *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i;

    for (i = rt->cols - 1; i >= rt->ccol + n; i--) {
        rt->cells[rt->crow][i] = rt->cells[rt->crow][i - n];
    }

    for (i = rt->ccol; i < rt->ccol + n; i++) {
        rt->cells[rt->crow][i].s[0] = 0x20;
        rt->cells[rt->crow][i].len  = 1;
        rt->cells[rt->crow][i].attr = rt->curattr;
    }

    rt->line_dirty[rt->crow] = true;
}

/* Interpret the 'delete chars' sequence (DCH) */
static void interpret_csi_DCH(RoteTerm *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i;

    for (i = rt->ccol; i < rt->cols; i++) {
        if (i + n < rt->cols) {
            rt->cells[rt->crow][i] = rt->cells[rt->crow][i + n];
        } else {
            rt->cells[rt->crow][i].s[0] = 0x20;
            rt->cells[rt->crow][i].len  = 1;
            rt->cells[rt->crow][i].attr = rt->curattr;
        }
    }

    rt->line_dirty[rt->crow] = true;
}

/* Interpret an 'insert line' sequence (IL) */
static void interpret_csi_IL(RoteTerm *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i, j;

    for (i = rt->pd->scrollbottom; i >= rt->crow + n; i--) {
        memcpy(rt->cells[i], rt->cells[i - n], sizeof(RoteCell) * rt->cols);
    }

    for (i = rt->crow; i < rt->crow + n && i <= rt->pd->scrollbottom; i++) {
        rt->line_dirty[i] = true;
        for (j = 0; j < rt->cols; j++) {
            rt->cells[i][j].s[0] = 0x20;
            rt->cells[i][j].len  = 1;
            rt->cells[i][j].attr = rt->curattr;
        }
    }

}

/* Interpret a 'delete line' sequence (DL) */
static void interpret_csi_DL(RoteTerm *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i, j;

    for (i = rt->crow; i <= rt->pd->scrollbottom; i++) {
        rt->line_dirty[i] = true;
        if (i + n <= rt->pd->scrollbottom) {
            memcpy(rt->cells[i], rt->cells[i+n], sizeof(RoteCell) * rt->cols);
        } else {
            for (j = 0; j < rt->cols; j++) {
                rt->cells[i][j].s[0] = 0x20;
                rt->cells[i][j].len  = 1;
                rt->cells[i][j].attr = rt->curattr;
            }
        }
    }
}

/* Interpret an 'erase characters' (ECH) sequence */
static void interpret_csi_ECH(RoteTerm *rt, int param[], int pcount)
{
    int n = (pcount && param[0] > 0) ? param[0] : 1;
    int i;

    for (i = rt->ccol; i < rt->ccol + n && i < rt->cols; i++) {
        rt->cells[rt->crow][i].s[0] = 0x20;
        rt->cells[rt->crow][i].len  = 1;
        rt->cells[rt->crow][i].attr = rt->curattr;
    }

    rt->line_dirty[rt->crow] = true;
}

/* Interpret a 'set scrolling region' (DECSTBM) sequence */
static void interpret_csi_DECSTBM(RoteTerm *rt, int param[], int pcount)
{
    int newtop, newbottom;

    if (!pcount) {
        newtop = 0;
        newbottom = rt->rows - 1;
    } else
    if (pcount < 2) {
        return; /* malformed */
    }

    newtop = param[0] - 1;
    newbottom = param[1] - 1;

    /* clamp to bounds */
    if (newtop < 0)
        newtop = 0;
    if (newtop >= rt->rows)
        newtop = rt->rows - 1;
    if (newbottom < 0)
        newbottom = 0;
    if (newbottom >= rt->rows)
        newbottom = rt->rows - 1;

    /* check for range validity */
    if (newtop > newbottom)
        return;
    rt->pd->scrolltop = newtop;
    rt->pd->scrollbottom = newbottom;
}

static void 
interpret_csi_SAVECUR(RoteTerm *rt,
                      int param[] __attribute__((unused)),
                      int pcount __attribute__((unused)))
{
    rt->pd->saved_x = rt->ccol;
    rt->pd->saved_y = rt->crow;
}

static void
interpret_csi_RESTORECUR(RoteTerm *rt,
                         int param[] __attribute__((unused)),
                         int pcount __attribute__((unused)))
{
    rt->ccol = rt->pd->saved_x;
    rt->crow = rt->pd->saved_y;
    rt->curpos_dirty = true;
}

static void rote_es_interpret_csi(RoteTerm *rt)
{
    static int csiparam[MAX_CSI_ES_PARAMS];
    int param_count = 0;
    const char *p = rt->pd->esbuf + 1;
    char verb = rt->pd->esbuf[rt->pd->esbuf_len - 1];

    if (!strncmp(rt->pd->esbuf, "[?", 2)) { /* private-mode CSI, ignore */
        return;
    }

    /* parse numeric parameters */
    while (isdigit((unsigned char)*p) || *p == ';') {
        if (*p == ';') {
            if (param_count >= MAX_CSI_ES_PARAMS) return; /* too long! */
            csiparam[param_count++] = 0;
        } else {
            if (param_count == 0) csiparam[param_count++] = 0;
            csiparam[param_count - 1] *= 10;
            csiparam[param_count - 1] += *p - '0';
        }

        p++;
    }

    /* delegate handling depending on command character (verb) */
    switch (verb) {
      case 'h':
        if (param_count == 1 && csiparam[0] == 4) /* insert mode */
            rt->insert = true;
        break;
      case 'l':
        if (param_count == 1 && csiparam[0] == 4) /* replace mode */
            rt->insert = false;
        break;
      case 'm': /* it's a 'set attribute' sequence */
        interpret_csi_SGR(rt, csiparam, param_count); break;
      case 'J': /* it's an 'erase display' sequence */
        interpret_csi_ED(rt, csiparam, param_count); break;
      case 'H': case 'f': /* it's a 'move cursor' sequence */
        interpret_csi_CUP(rt, csiparam, param_count); break;
      case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
      case 'e': case 'a': case 'd': case '`':
        /* it is a 'relative move' */
        interpret_csi_C(rt, verb, csiparam, param_count); break;
      case 'K': /* erase line */
        interpret_csi_EL(rt, csiparam, param_count); break;
      case '@': /* insert characters */
        interpret_csi_ICH(rt, csiparam, param_count); break;
      case 'P': /* delete characters */
        interpret_csi_DCH(rt, csiparam, param_count); break;
      case 'L': /* insert lines */
        interpret_csi_IL(rt, csiparam, param_count); break;
      case 'M': /* delete lines */
        interpret_csi_DL(rt, csiparam, param_count); break;
      case 'X': /* erase chars */
        interpret_csi_ECH(rt, csiparam, param_count); break;
      case 'r': /* set scrolling region */
        interpret_csi_DECSTBM(rt, csiparam, param_count); break;
      case 's': /* save cursor location */
        interpret_csi_SAVECUR(rt, csiparam, param_count); break;
      case 'u': /* restore cursor location */
        interpret_csi_RESTORECUR(rt, csiparam, param_count); break;
      default:
        break;
    }
}

RoteTerm *rote_vt_create(int rows, int cols)
{
    RoteTerm *rt;
    int i;

    if (rows <= 0 || cols <= 0)
        return NULL;

    rt = (RoteTerm*)calloc(sizeof(RoteTerm), 1);
    if (!rt)
        return NULL;

    /* record dimensions */
    rt->rows = rows;
    rt->cols = cols;

    /* default mode is replace */
    rt->insert = false; 

    /* create the cell matrix */
    rt->cells = (RoteCell**)calloc(sizeof(RoteCell*), rt->rows);
    for (i = 0; i < rt->rows; i++) {
        rt->cells[i] = (RoteCell*)calloc(sizeof(RoteCell), rt->cols);
    }

    /* allocate dirtiness array */
    rt->line_dirty = (bool*)calloc(sizeof(bool), rt->rows);

    /* initialization of other public fields */
    rt->crow = rt->ccol = 0;
    rt->curattr = 0x70;  /* white text over black background */

    /* allocate private data */
    rt->pd = (RoteTermPrivate*)calloc(sizeof(RoteTermPrivate), 1);

    rt->pty = -1;  /* no pty for now */

    /* initial scrolling area is the whole window */
    rt->pd->scrolltop = 0;
    rt->pd->scrollbottom = rt->rows - 1;

    return rt;
}

void rote_vt_destroy(RoteTerm *rt)
{
    int i;
    if (!rt)
        return;

    free(rt->pd);
    free(rt->line_dirty);
    for (i = 0; i < rt->rows; i++) {
        free(rt->cells[i]);
    }
    free(rt->cells);
    free(rt);
}

static void cur_set_attr(WINDOW *win, uint8_t attr)
{
    unsigned cp = ROTE_ATTR_BG(attr) * 8 + 7 - ROTE_ATTR_FG(attr);
    wattrset(win, cp ? COLOR_PAIR(cp) : A_NORMAL);

    if (ROTE_ATTR_BOLD(attr))
        wattron(win, A_BOLD);
    if (ROTE_ATTR_BLINK(attr))
        wattron(win, A_BLINK);
}

static unsigned ensure_printable(unsigned ch)
{
    return ch >= 32 ? ch : 32;
}

void rote_vt_draw(RoteTerm *rt, WINDOW *win, int srow, int scol)
{
    int i, j;

    for (i = 0; i < rt->rows; i++) {
        wmove(win, srow + i, scol);
        for (j = 0; j < rt->cols; j++) {
            cur_set_attr(win, rt->cells[i][j].attr);
            if (rt->cells[i][j].len && rt->cells[i][j].s[0] >= ' ') {
                waddnstr(win, rt->cells[i][j].s, rt->cells[i][j].len);
            } else {
                waddch(win, ' ');
            }
        }
    }

    wmove(win, srow + rt->crow, scol + rt->ccol);
}

/******************************************************/

pid_t rote_vt_forkpty(RoteTerm *rt, const char *path, const char *argv[])
{
    struct winsize ws;
    pid_t pid;

    ws.ws_row    = rt->rows;
    ws.ws_col    = rt->cols;
    ws.ws_xpixel = ws.ws_ypixel = 0;

    pid = forkpty(&rt->pty, NULL, NULL, &ws);
    if (pid < 0)
        return -1;

    if (pid == 0) {
        setsid();

        setenv("TERM", "linux", 1);
        execv(path, (char *const*)argv);
        fprintf(stderr, "\nexecv() failed.\nCommand: '%s'\n", argv[0]);
        exit(1);
    }

    return rt->childpid = pid;
}

int rote_vt_read(RoteTerm *rt, char *buf, int buflen)
{
    if (rt->pty < 0) {
        errno = EINVAL;
        return -1;
    }

    return read(rt->pty, buf, buflen);
}

int rote_vt_write(RoteTerm *rt, const char *data, int len)
{
    int res;

    if (rt->pty < 0) {
        errno = EINVAL;
        return -1;
    }

again:
    res = write(rt->pty, data, len);
    if (res < 0) {
        if (errno == EINTR || errno == EAGAIN)
            goto again;
    }

    return res;
}

static const char *keytable[KEY_MAX+1];

static void keytable_init()
{
    memset(keytable, 0, KEY_MAX+1 * sizeof(const char*));

    keytable['\n']          = "\r";
    keytable[KEY_UP]        = "\e[A";
    keytable[KEY_DOWN]      = "\e[B";
    keytable[KEY_RIGHT]     = "\e[C";
    keytable[KEY_LEFT]      = "\e[D";
    keytable[KEY_BACKSPACE] = "\b";
    keytable[KEY_HOME]      = "\e[1~";
    keytable[KEY_IC]        = "\e[2~";
    keytable[KEY_DC]        = "\e[3~";
    keytable[KEY_END]       = "\e[4~";
    keytable[KEY_PPAGE]     = "\e[5~";
    keytable[KEY_NPAGE]     = "\e[6~";
    keytable[KEY_SUSPEND]   = "\x1A";  /* Ctrl+Z gets mapped to this */
    keytable[KEY_F(1)]      = "\e[[A";
    keytable[KEY_F(2)]      = "\e[[B";
    keytable[KEY_F(3)]      = "\e[[C";
    keytable[KEY_F(4)]      = "\e[[D";
    keytable[KEY_F(5)]      = "\e[[E";
    keytable[KEY_F(6)]      = "\e[17~";
    keytable[KEY_F(7)]      = "\e[18~";
    keytable[KEY_F(8)]      = "\e[19~";
    keytable[KEY_F(9)]      = "\e[20~";
    keytable[KEY_F(10)]     = "\e[21~";
}

void rote_vt_keypress(RoteTerm *rt, int keycode)
{
    char c = (char)keycode;
    const char *buf;
    int len;

    if (keytable['\n'] == NULL)
        keytable_init();

    if (keycode >= 0 && keycode < KEY_MAX && keytable[keycode]) {
        buf = keytable[keycode];
        len = strlen(keytable[keycode]);
    } else {
        buf = &c;
        len = 1;
    }

    while (len > 0) {
        int res = rote_vt_write(rt, buf, len);
        if (res < 0)
            return;

        buf += res;
        len -= res;
    }
}

